using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;

namespace Chip8 {
    class MidiDevice {
        
        [DllImport("winmm.dll")]
        private static extern MidiError midiOutOpen(ref int lphMidiOut, int uDeviceID, int dwCallback, int dwInstance, int dwFlags);

        [DllImport("winmm.dll")]
        private static extern MidiError midiOutClose(int hMidiOut);

        [DllImport("winmm.dll")]
        private static extern MidiError midiOutShortMsg(int hMidiOut, int dwMsg);

        [DllImport("winmm.dll", EntryPoint = "midiOutGetDevCapsA")]
        private static extern MidiError midiOutGetDevCaps(int uDeviceID, ref MIDIOUTCAPS lpCaps, int uSize);

        private const int MAXPNAMELEN = 32;

        [StructLayout(LayoutKind.Sequential)]
        public struct MIDIOUTCAPS {
            internal Int16 wMid;
            internal Int16 wPid;
            internal int vDriverVersion;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = MAXPNAMELEN)]
            internal string szPname;
            internal Int16 wTechnology;
            internal Int16 wVoices;
            internal Int16 wNotes;
            internal Int16 wChannelMask;
            internal int dwSupport;
        }

        [DllImport("winmm.dll")]
        private static extern Int16 midiOutGetNumDevs();


        public enum MidiMessage {
            KeyRelease = 0x80,
            KeyPress = 0x90,
            KeyAfterTouch = 0xA0,
            ControlChange = 0xB0,
            PatchChange = 0xC0,
            ChannelAfterTouch = 0xD0,
            PitchWheelChange = 0xE0
        };



        private const int MMSYSERR_BASE = 0;
        private const int MIDIERR_BASE = 64;
        
        public enum MidiError {
            MMSYSERR_NOERROR = 0,
            MIDIERR_NODEVICE = (MIDIERR_BASE + 4),
            MMSYSERR_ALLOCATED = (MMSYSERR_BASE + 4),
            MMSYSERR_BADDEVICEID = (MMSYSERR_BASE + 2),
            MMSYSERR_INVALPARAM = (MMSYSERR_BASE + 11),
            MMSYSERR_NOMEM = (MMSYSERR_BASE + 7),
            MIDIERR_STILLPLAYING = (MIDIERR_BASE + 1),
            MMSYSERR_INVALHANDLE = (MMSYSERR_BASE + 5),
            MMSYSERR_BADERRNUM = (MMSYSERR_BASE + 9)
        };

        public void SendMidiMessage(MidiMessage Type, int Channel) {
            SendMidiMessage(Type, Channel, 0, 0);
        }
        public void SendMidiMessage(MidiMessage Type, int Channel, int DataByte1) {
            SendMidiMessage(Type, Channel, DataByte1, 0);
        }
        public void SendMidiMessage(MidiMessage Type, int Channel, int DataByte1, int DataByte2) {
            MidiError Result = midiOutShortMsg(MidiHandle, ((int)Type | Channel) | (DataByte1 << 8) | (DataByte2 << 16));
            if (Result != MidiError.MMSYSERR_NOERROR) throw new Exception("Could not send MIDI message: " + Result.ToString());
        }

        private int MidiHandle = -1;

        public int DeviceCount = -1;

        public readonly MIDIOUTCAPS[] Devices;

        public MidiDevice() {

            List<MIDIOUTCAPS> FindDevices = new List<MIDIOUTCAPS>();

            DeviceCount = midiOutGetNumDevs();
            Devices = new MIDIOUTCAPS[DeviceCount];

            for (int i = 0; i < DeviceCount; ++i) {
                Devices[i] = new MIDIOUTCAPS();
                midiOutGetDevCaps(i, ref Devices[i], Marshal.SizeOf(Devices[i]));
            }

            MidiError Result = midiOutOpen(ref this.MidiHandle, -1, 0, 0, 0);
            if (Result != MidiError.MMSYSERR_NOERROR) {
                throw new Exception("Opening MIDI device failed with error " + Result.ToString());
            }
        }

        ~MidiDevice() {
            MidiError Result = midiOutClose(this.MidiHandle);
            if (Result != MidiError.MMSYSERR_NOERROR) {
                throw new Exception("Closing MIDI device failed with error " + Result.ToString());
            }
        }
    }
}
